"""kernal functions for poly-to-poly area-weighted mapping."""
import copy
import logging
import multiprocessing
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
from pygeos import GEOSException

from .gdp_data_class import TiffAttributes
from .helpers import build_subset_tiff
from .utils import _get_crs
from .utils import _get_print_on
from .utils import _get_shp_bounds_w_buffer
from .utils import _quadrat_cut_geometry

logger = logging.getLogger(__name__)


def tiff_intersection(
    data: TiffAttributes,
    gdf: gpd.GeoDataFrame,
    poly_idx: str,
    wght_gen_proj: Any,
) -> gpd.GeoDataFrame:
    """Tiff intersection.

    Args:
        data (TiffAttributes): _description_
        gdf (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_
        wght_gen_proj (Any): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    feat_bounds = _get_shp_bounds_w_buffer(
        gdf=gdf, ds=data.ds, crs=data.crs, lon=data.xname, lat=data.yname
    )

    subset_dict = build_subset_tiff(
        bounds=feat_bounds,
        xname=data.xname,
        yname=data.yname,
        toptobottom=data.toptobottom,
        bname=data.bname,
        band=data.band,
    )
    ds_ss = data.ds.sel(**subset_dict).rio.interpolate_na()
    lon, lat = np.meshgrid(ds_ss[data.xname].values, ds_ss[data.yname].values)
    lat_flat = lat.flatten()
    lon_flat = lon.flatten()
    df_points = pd.DataFrame(
        {"index": np.arange(len(lat_flat)), "lat": lat_flat, "lon": lon_flat}
    )
    dm_pnt_gdf = gpd.GeoDataFrame(
        df_points, geometry=gpd.points_from_xy(df_points.lon, df_points.lat)
    )
    dm_pnt_gdf.set_crs(data.crs, inplace=True)
    wproj = _get_crs(wght_gen_proj)
    dm_pnt_gdf_wproj = dm_pnt_gdf.to_crs(wproj)
    gdf_wproj = gdf.to_crs(wproj)

    # pintersect = gpd.sjoin(dm_pnt_gdf_wproj, gdf_wproj, op='within', how='right')
    pintersect = dm_pnt_gdf_wproj.sjoin(gdf_wproj, how="right", predicate="intersects")
    # pintersect_nr = gpd.sjoin_nearest(dm_pnt_gdf_wproj, gdf_wproj, how='right')
    pintersect_nr = dm_pnt_gdf_wproj.sjoin_nearest(gdf_wproj, how="right")

    ds_vals = ds_ss.values.flatten()
    if data.categorical:
        d_categories = list(pd.Categorical(ds_vals).categories)
    num_features = len(gdf.index)
    u_ids = pintersect.groupby(poly_idx, dropna=False)
    u_ids_nr = pintersect_nr.groupby(poly_idx, dropna=False)
    for index in range(num_features):
        feature = gdf.iloc[[index]]
        feature_id = feature[poly_idx].values[0]
        tiff_inds_nr = u_ids_nr.get_group(feature_id).index_left.values
        tiff_inds = u_ids.get_group(feature_id).index_left.values
        if not np.isnan(tiff_inds).any():
            tval = (
                pd.Series(ds_vals[tiff_inds.astype(int)], dtype="category").to_frame()
                if data.categorical
                else pd.Series(ds_vals[tiff_inds.astype(int)]).to_frame()
            )
        else:
            tval = (
                pd.Series(
                    ds_vals[tiff_inds_nr.astype(int)], dtype="category"
                ).to_frame()
                if data.categorical
                else pd.Series(ds_vals[tiff_inds_nr.astype(int)]).to_frame()
            )

        if data.categorical:
            stats = tval.describe(include=["category"]).T
            stats_freq = (
                pd.Categorical(ds_vals[tiff_inds.astype(int)], categories=d_categories)
                .describe()
                .freqs.T
            )
            stats = stats.merge(stats_freq, how="cross")
        else:
            stats = tval.describe().T
        # stats.insert(0, poly_idx, gdf[poly_idx])
        statframe: pd.DataFrame = (
            stats if index == 0 else pd.concat([statframe, stats], axis=0)
        )
    statframe.insert(0, poly_idx, gdf[poly_idx].values)
    statframe.set_index(poly_idx, inplace=True)

    return statframe


def chunks(
    data: TiffAttributes,
    gdf: gpd.GeoDataFrame,
    poly_idx: str,
    wght_gen_proj: Any,
    num_cores: int,
) -> Tuple[TiffAttributes, gpd.GeoDataFrame, str, Any]:
    """Chuck gdf for multiprocessing.

    Args:
        data (TiffAttributes): _description_
        gdf (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_
        wght_gen_proj (Any): _description_
        num_cores (int): _description_

    Yields:
        Iterator[Tuple[TiffAttributes, gpd.GeoDataFrame, str, Any]]: _description_
    """
    num_features = len(gdf.index)
    for i in range(0, num_features, num_cores):
        yield (
            copy.deepcopy(data),
            gdf.iloc[i : i + num_cores],
            poly_idx,
            wght_gen_proj,
        )


# This doesn't work because I think of collisions when subsetting the xarray.
# TODO: sounds like I need to pass the path to file, and have each process read file.
def p_tiff_intersection_by_feature(
    data: TiffAttributes,
    gdf: gpd.GeoDataFrame,
    poly_idx: str,
    wght_gen_proj: Any,
) -> gpd.GeoDataFrame:
    """Multiprocessing version of tiff_intersection_by_feature.

    Args:
        data (TiffAttributes): _description_
        gdf (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_
        wght_gen_proj (Any): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    cores = multiprocessing.cpu_count()
    p = multiprocessing.Pool(cores - 1)
    return p.starmap(
        tiff_intersection_by_feature,
        chunks(
            data=data,
            gdf=gdf,
            poly_idx=poly_idx,
            wght_gen_proj=wght_gen_proj,
            num_cores=cores - 1,
        ),
    )


def tiff_intersection_by_feature(
    data: TiffAttributes,
    gdf: gpd.GeoDataFrame,
    poly_idx: str,
) -> pd.DataFrame:
    """Tiff intersection.

    Args:
        data (TiffAttributes): _description_
        gdf (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    num_features = len(gdf.index)
    print_val = _get_print_on(num_features)
    wproj = _get_crs(data.crs)
    gdf_wproj = gdf.to_crs(wproj)
    if data.categorical:
        d_categories = list(pd.Categorical(data.ds.values.flatten()).categories)

    for index in range(num_features):
        if index % print_val == 0:
            print(f"Processing feature: {index}")
        feature = gdf_wproj.iloc[[index]]
        if data.categorical:
            stats = _tiff_get_stat_by_feature_2(
                data=data,
                feature=feature,
                feat_index=index,
                poly_idx=poly_idx,
                d_cats=d_categories,
            )
        else:
            stats = _tiff_get_stat_by_feature_2(
                data=data, feature=feature, feat_index=index, poly_idx=poly_idx
            )
        statframe: pd.DataFrame = (
            stats if index == 0 else pd.concat([statframe, stats], axis=0)
        )
    # statframe.insert(0, poly_idx, gdf_wproj[poly_idx].values)
    statframe.set_index(poly_idx, inplace=True)

    return statframe


def _tiff_get_stat_by_feature(
    data: TiffAttributes,
    feature: gpd.GeoDataFrame,
    poly_idx: str,
    d_cats: Optional[List[object]] = None,
) -> pd.DataFrame:
    """Generate stats per feature.

    Args:
        data (TiffAttributes): _description_
        feature (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_
        d_cats (Optional[List[object]], optional): _description_. Defaults to None.

    Returns:
        pd.DataFrame: _description_
    """
    feat_bounds = _get_shp_bounds_w_buffer(
        gdf=feature, ds=data.ds, crs=data.crs, lon=data.xname, lat=data.yname
    )

    subset_dict = build_subset_tiff(
        bounds=feat_bounds,
        xname=data.xname,
        yname=data.yname,
        toptobottom=data.toptobottom,
        bname=data.bname,
        band=data.band,
    )
    ds_ss = data.ds.sel(**subset_dict).rio.interpolate_na()
    lon, lat = np.meshgrid(ds_ss[data.xname].values, ds_ss[data.yname].values)
    lat_flat = lat.flatten()
    lon_flat = lon.flatten()
    df_points = pd.DataFrame(
        {"index": np.arange(len(lat_flat)), "lat": lat_flat, "lon": lon_flat}
    )
    dm_pnt_gdf = gpd.GeoDataFrame(
        df_points, geometry=gpd.points_from_xy(df_points.lon, df_points.lat)
    )
    dm_pnt_gdf.set_crs(data.crs, inplace=True)
    dm_pnt_gdf_wproj = dm_pnt_gdf.to_crs(data.crs)

    # pintersect = gpd.sjoin(dm_pnt_gdf_wproj, gdf_wproj, op='within', how='right')
    pintersect = dm_pnt_gdf_wproj.sjoin(feature, how="right", predicate="intersects")
    # pintersect_nr = gpd.sjoin_nearest(dm_pnt_gdf_wproj, gdf_wproj, how='right')
    pintersect_nr = dm_pnt_gdf_wproj.sjoin_nearest(feature, how="right")

    ds_vals = ds_ss.values.flatten()

    u_ids = pintersect.groupby(poly_idx, dropna=False)
    u_ids_nr = pintersect_nr.groupby(poly_idx, dropna=False)

    feature_id = feature[poly_idx].values[0]
    tiff_inds_nr = u_ids_nr.get_group(feature_id).index_left.values
    tiff_inds = u_ids.get_group(feature_id).index_left.values
    if not np.isnan(tiff_inds).any():
        tval = (
            pd.Series(ds_vals[tiff_inds.astype(int)], dtype="category").to_frame()
            if data.categorical
            else pd.Series(ds_vals[tiff_inds.astype(int)]).to_frame()
        )
    else:
        tval = (
            pd.Series(ds_vals[tiff_inds_nr.astype(int)], dtype="category").to_frame()
            if data.categorical
            else pd.Series(ds_vals[tiff_inds_nr.astype(int)]).to_frame()
        )
    # elif data.categorical:
    #     tval = pd.Series(ds_vals[tiff_inds_nr.astype(int)], dtype="category").to_frame()
    # else:
    #     tval = pd.Series(ds_vals[tiff_inds_nr.astype(int)]).to_frame()
    if data.categorical:
        stats = tval.describe(include=["category"]).T
        if not np.isnan(tiff_inds).any():
            stats_freq = (
                pd.Categorical(ds_vals[tiff_inds.astype(int)], categories=d_cats)
                .describe()
                .T
            )
        else:
            stats_freq = (
                pd.Categorical(ds_vals[tiff_inds_nr.astype(int)], categories=d_cats)
                .describe()
                .T
            )
        stats = stats.merge(stats_freq.iloc[[1]], how="cross")
    else:
        stats = tval.describe().T
    stats.insert(0, poly_idx, feature[poly_idx].values[0])
    return stats


def _tiff_get_stat_by_feature_2(  # noqa:C901
    data: TiffAttributes,
    feature: gpd.GeoDataFrame,
    feat_index: int,
    poly_idx: str,
    d_cats: Optional[List[object]] = None,
) -> pd.DataFrame:
    """Generate stats per feature.

    Args:
        data (TiffAttributes): _description_
        feature (gpd.GeoDataFrame): _description_
        feat_index (int): _description_
        poly_idx (str): _description_
        d_cats (Optional[List[object]], optional): _description_. Defaults to None.

    Returns:
        pd.DataFrame: _description_
    """
    feat_bounds = _get_shp_bounds_w_buffer(
        gdf=feature, ds=data.ds, crs=data.crs, lon=data.xname, lat=data.yname
    )

    subset_dict = build_subset_tiff(
        bounds=feat_bounds,
        xname=data.xname,
        yname=data.yname,
        toptobottom=data.toptobottom,
        bname=data.bname,
        band=data.band,
    )
    ds_ss = data.ds.sel(**subset_dict).rio.interpolate_na()
    lon, lat = np.meshgrid(ds_ss[data.xname].values, ds_ss[data.yname].values)
    lat_flat = lat.flatten()
    lon_flat = lon.flatten()
    df_points = pd.DataFrame(
        {"index": np.arange(len(lat_flat)), "lat": lat_flat, "lon": lon_flat}
    )
    dm_pnt_gdf = gpd.GeoDataFrame(
        df_points, geometry=gpd.points_from_xy(df_points.lon, df_points.lat)
    )
    dm_pnt_gdf.set_crs(data.crs, inplace=True)
    dm_pnt_gdf_wproj = dm_pnt_gdf.to_crs(data.crs)
    dx, dy = data.ds.rio.resolution()
    multipoly = _quadrat_cut_geometry(
        feature["geometry"][feat_index], np.mean([abs(dx), abs(dy)]) * 50.0
    )
    spatial_index = dm_pnt_gdf_wproj.sindex
    ds_vals = ds_ss.values.flatten()
    geoms = set()
    for spindex, poly in enumerate(multipoly.geoms):
        try:
            possible_matches_index = list(spatial_index.intersection(poly.bounds))
        except AttributeError:
            print(
                f"User feature Attribute error index: {feat_index} \
                    subpoly_index: {spindex} has an error"
            )

        if possible_matches_index:
            possible_matches = dm_pnt_gdf_wproj.iloc[possible_matches_index]
            try:
                precise_matches = possible_matches[possible_matches.intersects(poly)]
            except GEOSException:
                print(
                    f"User feature GEOSException error: index={feat_index}, \
                    sub_poly={spindex}"
                )
            except TypeError:
                print(
                    f"User feature Type error: index={feat_index}, \
                    sub_poly={spindex}"
                )
            geoms.update(precise_matches.index)

    if list(geoms):
        tval = (
            pd.Series(ds_vals[list(geoms)], dtype="category").to_frame()
            if data.categorical
            else pd.Series(ds_vals[list(geoms)]).to_frame()
        )
    else:
        nrpnt = spatial_index.nearest(feature["geometry"].centroid, return_all=False)
        tval = (
            pd.Series(ds_vals[nrpnt[1, :]], dtype="category").to_frame()
            if data.categorical
            else pd.Series(ds_vals[nrpnt[1, :]]).to_frame()
        )

    if data.categorical:
        stats = tval.describe(include=["category"]).T
        if not np.isnan(list(geoms)).any():
            stats_freq = (
                pd.Categorical(ds_vals[list(geoms)], categories=d_cats).describe().T
            )
        else:
            stats_freq = (
                pd.Categorical(ds_vals[nrpnt[1, :]], categories=d_cats).describe().T
            )
        stats = stats.merge(stats_freq.iloc[[1]], how="cross")
    else:
        stats = tval.describe().T
        stats.insert(0, "sum", tval.sum())
    stats.insert(0, poly_idx, feature[poly_idx].values[0])
    return stats


def tiff_intersection_by_feature_2(
    data: TiffAttributes,
    gdf: gpd.GeoDataFrame,
    poly_idx: str,
) -> gpd.GeoDataFrame:
    """Tiff intersection.

    Args:
        data (TiffAttributes): _description_
        gdf (gpd.GeoDataFrame): _description_
        poly_idx (str): _description_

    Returns:
        gpd.GeoDataFrame: _description_
    """
    num_features = len(gdf.index)
    print_val = _get_print_on(num_features)
    wproj = _get_crs(data.crs)
    gdf_wproj = gdf.to_crs(wproj)
    if data.categorical:
        d_categories = list(pd.Categorical(data.ds.values.flatten()).categories)

    for index in range(num_features):
        if index % print_val == 0:
            print(f"Processing feature: {index}")
        feature = gdf_wproj.iloc[[index]]
        if data.categorical:
            stats = _tiff_get_stat_by_feature(
                data=data,
                feature=feature,
                poly_idx=poly_idx,
                d_cats=d_categories,
            )
        else:
            stats = _tiff_get_stat_by_feature(
                data=data,
                feature=feature,
                feat_index=index,
                poly_idx=poly_idx,
                wproj=wproj,
            )
        statframe: pd.DataFrame = (
            stats if index == 0 else pd.concat([statframe, stats], axis=0)
        )
    # statframe.insert(0, poly_idx, gdf_wproj[poly_idx].values)
    statframe.set_index(poly_idx, inplace=True)

    return statframe



"""Abstract Base Class for Template behavior pattern for calculating weights."""
import os
import time
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
import pygeos
import pyproj
from joblib import delayed
from joblib import Parallel
from joblib import parallel_backend

from gdptools.utils import _check_feature_crs
from gdptools.utils import _check_grid_cell_crs
from gdptools.utils import _check_poly_idx
from gdptools.utils import _get_crs
from gdptools.utils import _reproject_for_weight_calc


class CalcWeightEngine(ABC):
    """Abstract Base Class (ABC) implementing the template behavioral pattern.

    Abstract Base Class for calculating weights.  There are several weight generation
    methods implemented and they all share a common workflow with different methods
    for calculating the weighs.  This ABC defines the calc_weights() workflow, with
    an @abstractmethod for get_weight_components() where new methods can be plugged
    in for weight generation.
    """

    def calc_weights(
        self,
        poly: gpd.GeoDataFrame,
        poly_idx: str,
        grid_cells: gpd.GeoDataFrame,
        wght_gen_crs: Any,
        filename: Optional[str] = "",
        intersections: Optional[bool] = False,
        verbose: Optional[bool] = False,
    ) -> Tuple[pd.DataFrame, Optional[gpd.GeoDataFrame]]:
        """Template method for calculating weights.

        Args:
            poly (gpd.GeoDataFrame): _description_
            poly_idx (str): _description_
            grid_cells (gpd.GeoDataFrame): _description_
            wght_gen_crs (Any): _description_
            filename (Optional[str], optional): _description_. Defaults to "".
            intersections (Optional[bool], optional): _description_. Defaults to False.
            verbose (Optional[bool], optional): _description_. Defaults to False.

        Returns:
            Tuple[pd.DataFrame, Optional[gpd.GeoDataFrame]]: _description_
        """
        self.poly = poly.copy()
        self.poly_idx = poly_idx
        self.grid_cells = grid_cells.copy()
        self.wght_gen_crs = wght_gen_crs
        self.filename = filename
        self.intersections = intersections
        self.verbose = verbose
        self.check_poly_idx()
        self.check_feature_crs()
        self.check_grid_cell_crs()
        self.grid_out_crs = self.get_crs()
        self.reproject_for_weight_calc()
        if not self.intersections:
            (
                self.plist,
                self.ilist,
                self.jlist,
                self.wghtlist,
            ) = self.get_weight_components()
        else:
            print(f"Intersections = {self.intersections}")
            (
                self.plist,
                self.ilist,
                self.jlist,
                self.wghtlist,
                self.calc_intersect,
            ) = self.get_weight_components_and_intesections()
        self.wght_df = self.create_wght_df()
        if self.filename:
            self.wght_df.to_csv(self.filename)

        return (
            (self.wght_df, self.calc_intersect) if self.intersections else self.wght_df
        )

    def check_poly_idx(self):
        """Check poly_idx in poly GeoDataFrame."""
        _check_poly_idx(self.poly, self.poly_idx)

    def check_feature_crs(self):
        """Check poly contains crs."""
        _check_feature_crs(self.poly)

    def check_grid_cell_crs(self):
        """Check grid_cells contains crs."""
        _check_grid_cell_crs

    def get_crs(self) -> pyproj.CRS:
        """Check/get crs of wght_gen_crs."""
        return _get_crs(self.wght_gen_crs)

    def reproject_for_weight_calc(self):
        """Reproject poly and grid_cells into wght_gen_crs."""
        _reproject_for_weight_calc(
            self.poly, self.grid_cells, self.wght_gen_crs, self.grid_out_crs
        )

    @abstractmethod
    def get_weight_components(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float]]:
        """Abstract method for calculating weights.

        Classes that inherit this method will override this method for \
            weight-generation.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        pass

    @abstractmethod
    def get_weight_components_and_intesections(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float], gpd.GeoDataFrame]:
        """Abstract method for calculating weights.

        Classes that inherit this method will override this method for \
            weight-generation.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        pass

    def create_wght_df(self) -> pd.DataFrame:
        """Create dataframe from weight components."""
        wght_df = pd.DataFrame(
            {
                self.poly_idx: self.plist,
                "i": self.ilist,
                "j": self.jlist,
                "wght": self.wghtlist,
            }
        )
        wght_df = wght_df.astype(
            {"i": int, "j": int, "wght": float, self.poly_idx: str}
        )
        return wght_df


class SerialWghtGenEngine(CalcWeightEngine):
    """Tobbler python package Method to generate grid-to-polygon weight.

    This class is based on methods provided in the Tobbler package. See
        area_tables_bining() method.

    Args:
        CalcWeightEngine (ABC): Abstract Base Class (ABC) employing the Template behavior
            pattern.  The abstractmethod get weight components povides a method to plug-
            in new weight generation methods.
    """

    def get_weight_components(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float]]:
        """Template method from CalcWeightEngine class for generating weight components.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tsrt = time.perf_counter()
        (
            plist,
            ilist,
            jlist,
            wghtslist,
        ) = self.area_tables_binning(source_df=self.grid_cells, target_df=self.poly)
        tend = time.perf_counter()
        print(f"Wight gen finished in {tend-tsrt:0.4f} seconds")

        return plist, ilist, jlist, wghtslist

    def get_weight_components_and_intesections(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float], gpd.GeoDataFrame]:
        """Template method from CalcWeightEngine class for generating weight components.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tsrt = time.perf_counter()
        (
            plist,
            ilist,
            jlist,
            wghtslist,
            gdf,
        ) = self.area_tables_binning_and_intersections(
            source_df=self.grid_cells, target_df=self.poly
        )
        tend = time.perf_counter()
        print(f"Wight gen finished in {tend-tsrt:0.4f} seconds")

        return plist, ilist, jlist, wghtslist, gdf

    def area_tables_binning(
        self: "SerialWghtGenEngine",
        source_df: gpd.GeoDataFrame,
        target_df: gpd.GeoDataFrame,
    ) -> Tuple[List[str], List[int], List[int], List[float]]:
        """Construct intersection tables.

        Construct area allocation and source-target correspondence tables using
        a parallel spatial indexing approach. This method and its associated functions
        are based on and adapted from the Tobbler package:

        https://github.com/pysal/tobler.

        Tobler License:
        BSD 3-Clause License

        Copyright 2018 pysal-spopt developers

        Redistribution and use in source and binary forms, with or without modification,
            are permitted provided that the following conditions are met:

        1.  Redistributions of source code must retain the above copyright notice, this
            list of conditions and the following disclaimer.

        2.  Redistributions in binary form must reproduce the above copyright notice,
            this list of conditions and the following disclaimer in the documentation
            and/or other materials provided with the distribution.

        3.  Neither the name of the copyright holder nor the names of its contributors
            may be used to endorse or promote products derived from this software
            without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
        (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
        LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
        ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        Args:
            source_df (gpd.GeoDataFrame): GeoDataFrame containing input data and
                polygons
            target_df (gpd.GeoDataFrame): GeoDataFrame defining the output geometries

        Returns:
            Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tstrt = time.perf_counter()
        df1 = _make_valid(source_df)
        df2 = _make_valid(target_df)
        tend = time.perf_counter()
        print(f"Validate polygons finished in {tend - tstrt:0.4f} seconds")

        tstrt = time.perf_counter()
        ids_tgt, ids_src = df1.sindex.query_bulk(df2.geometry, predicate="intersects")

        areas = (
            df1.geometry.values[ids_src].intersection(df2.geometry.values[ids_tgt]).area
            / df2.geometry.values[ids_tgt].area
        )
        tend = time.perf_counter()
        print(f"Intersections finished in {tend - tstrt:0.4f} seconds")

        return (
            df2[self.poly_idx].iloc[ids_tgt].values.astype(object).tolist(),
            df1.i_index.iloc[ids_src].values.astype(int).tolist(),
            df1.j_index.iloc[ids_src].values.astype(int).tolist(),
            areas.astype(float).tolist(),
        )

    def area_tables_binning_and_intersections(
        self: "SerialWghtGenEngine",
        source_df: gpd.GeoDataFrame,
        target_df: gpd.GeoDataFrame,
    ) -> Tuple[List[str], List[int], List[int], List[float], gpd.GeoDataFrame]:
        """Construct intersection tables.

        Construct area allocation and source-target correspondence tables using
        a parallel spatial indexing approach. This method and its associated functions
        are based on and adapted from the Tobbler package:

        https://github.com/pysal/tobler.

        Tobler License:
        BSD 3-Clause License

        Copyright 2018 pysal-spopt developers

        Redistribution and use in source and binary forms, with or without modification,
            are permitted provided that the following conditions are met:

        1.  Redistributions of source code must retain the above copyright notice, this
            list of conditions and the following disclaimer.

        2.  Redistributions in binary form must reproduce the above copyright notice,
            this list of conditions and the following disclaimer in the documentation
            and/or other materials provided with the distribution.

        3.  Neither the name of the copyright holder nor the names of its contributors
            may be used to endorse or promote products derived from this software
            without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
        (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
        LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
        ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        Args:
            source_df (gpd.GeoDataFrame): GeoDataFrame containing input data and
                polygons
            target_df (gpd.GeoDataFrame): GeoDataFrame defining the output geometries

        Returns:
            Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tstrt = time.perf_counter()
        df1 = _make_valid(source_df)
        df2 = _make_valid(target_df)
        tend = time.perf_counter()
        print(f"Validate polygons finished in {tend - tstrt:0.4f} seconds")

        tstrt = time.perf_counter()
        ids_tgt, ids_src = df1.sindex.query_bulk(df2.geometry, predicate="intersects")
        f_intersect = df1.geometry.values[ids_src].intersection(
            df2.geometry.values[ids_tgt]
        )
        areas = f_intersect.area / df2.geometry.values[ids_tgt].area
        gdf_inter = df2.iloc[ids_tgt]
        gdf_inter.set_geometry(f_intersect, inplace=True)
        tend = time.perf_counter()
        print(f"Intersections finished in {tend - tstrt:0.4f} seconds")

        return (
            df2[self.poly_idx].iloc[ids_tgt].values.astype(object).tolist(),
            df1.i_index.iloc[ids_src].values.astype(int).tolist(),
            df1.j_index.iloc[ids_src].values.astype(int).tolist(),
            areas.astype(float).tolist(),
            gdf_inter,
        )


class ParallelWghtGenEngine(CalcWeightEngine):
    """Tobbler package Method to generate grid-to-polygon weight.

    This class is based on methods provided in the Tobbler package. See
        area_tables_bining_parallel() method.

    Args:
        CalcWeightEngine (ABC): Abstract Base Class (ABC) employing the Template behavior
            pattern.  The abstractmethod get weight components povides a method to plug-
            in new weight generation methods.
    """

    def get_weight_components(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float]]:
        """Template method from CalcWeightEngine class for generating weight components.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tsrt = time.perf_counter()
        plist, ilist, jlist, wghtslist = self.area_tables_binning_parallel(
            source_df=self.grid_cells, target_df=self.poly
        )
        tend = time.perf_counter()
        print(f"Wight gen finished in {tend-tsrt:0.4f} seconds")

        return plist, ilist, jlist, wghtslist

    def get_weight_components_and_intesections(
        self,
    ) -> Tuple[List[str], List[int], List[int], List[float], gpd.GeoDataFrame]:
        """Template method from CalcWeightEngine class for generating weight components.

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        tsrt = time.perf_counter()
        (
            plist,
            ilist,
            jlist,
            wghtslist,
            gdf,
        ) = self.area_tables_binning_parallel_and_intersections(
            source_df=self.grid_cells, target_df=self.poly
        )
        tend = time.perf_counter()
        print(f"Wight gen finished in {tend-tsrt:0.4f} seconds")

        return plist, ilist, jlist, wghtslist, gdf

    def area_tables_binning_parallel(
        self,
        source_df: gpd.GeoDataFrame,
        target_df: gpd.GeoDataFrame,
        n_jobs: Optional[int] = -1,
    ) -> Tuple[List[str], List[int], List[int], List[float]]:
        """Construct intersection tables.

        Construct area allocation and source-target correspondence tables using
        a parallel spatial indexing approach. This method and its associated functions
        are based on and adapted from the Tobbler package:

        https://github.com/pysal/tobler.

        Tobler License:
        BSD 3-Clause License

        Copyright 2018 pysal-spopt developers

        Redistribution and use in source and binary forms, with or without modification,
            are permitted provided that the following conditions are met:

        1.  Redistributions of source code must retain the above copyright notice, this
            list of conditions and the following disclaimer.

        2.  Redistributions in binary form must reproduce the above copyright notice,
            this list of conditions and the following disclaimer in the documentation
            and/or other materials provided with the distribution.

        3.  Neither the name of the copyright holder nor the names of its contributors
            may be used to endorse or promote products derived from this software
            without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
        (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
        LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
        ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        Args:
            source_df (gpd.GeoDataFrame): GeoDataFrame containing input data and
                polygons
            target_df (gpd.GeoDataFrame): GeoDataFrame defining the output geometries
            n_jobs (int, optional): [Optional. Default=-1]
                Number of processes to run in parallel. If -1, this is set to the
                number of CPUs available

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        if n_jobs == -1:
            n_jobs = os.cpu_count()

        # Buffer polygons with self-intersections
        tstrt = time.perf_counter()
        df1 = _make_valid(source_df)
        df2 = _make_valid(target_df)
        tend = time.perf_counter()
        print(f"Validate polygons finished in {tend - tstrt:0.4f} seconds")

        # Chunk the largest, ship the smallest in full
        to_chunk, df_full = self.get_chunks(df1, df2)

        # Spatial index query: Reindex on positional IDs
        to_workers = _chunk_dfs(
            gpd.GeoSeries(to_chunk.geometry.values, crs=to_chunk.crs),
            gpd.GeoSeries(df_full.geometry.values, crs=df_full.crs),
            n_jobs,
        )

        worker_out = self.get_ids_parallel(n_jobs, to_workers)
        ids_src, ids_tgt = np.concatenate(worker_out).T

        # Intersection + area calculation
        chunks_to_intersection = _chunk_polys(
            np.vstack([ids_src, ids_tgt]).T, df1.geometry, df2.geometry, n_jobs
        )
        worker_out = self.get_areas_parallel(n_jobs, chunks_to_intersection)
        areas = np.concatenate(worker_out)

        return (
            df2[self.poly_idx].iloc[ids_tgt].values.astype(object).tolist(),
            df1.i_index.iloc[ids_src].values.astype(int).tolist(),
            df1.j_index.iloc[ids_src].values.astype(int).tolist(),
            (areas.astype(float) / df2.geometry[ids_tgt].area).tolist(),
        )

    def area_tables_binning_parallel_and_intersections(
        self,
        source_df: gpd.GeoDataFrame,
        target_df: gpd.GeoDataFrame,
        n_jobs: Optional[int] = -1,
    ) -> Tuple[List[str], List[int], List[int], List[float], gpd.GeoDataFrame]:
        """Construct intersection tables.

        Construct area allocation and source-target correspondence tables using
        a parallel spatial indexing approach. This method and its associated functions
        are based on and adapted from the Tobbler package:

        https://github.com/pysal/tobler.

        Tobler License:
        BSD 3-Clause License

        Copyright 2018 pysal-spopt developers

        Redistribution and use in source and binary forms, with or without modification,
            are permitted provided that the following conditions are met:

        1.  Redistributions of source code must retain the above copyright notice, this
            list of conditions and the following disclaimer.

        2.  Redistributions in binary form must reproduce the above copyright notice,
            this list of conditions and the following disclaimer in the documentation
            and/or other materials provided with the distribution.

        3.  Neither the name of the copyright holder nor the names of its contributors
            may be used to endorse or promote products derived from this software
            without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
        (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
        LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
        ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        Args:
            source_df (gpd.GeoDataFrame): GeoDataFrame containing input data and
                polygons
            target_df (gpd.GeoDataFrame): GeoDataFrame defining the output geometries
            n_jobs (int, optional): [Optional. Default=-1]
                Number of processes to run in parallel. If -1, this is set to the
                number of CPUs available

        Returns:
            Tuple[List[str], List[int], List[int], List[float]]:
            Returned tuples in order:
                1) plist: list of poly_idx strings.
                2) ilist i-index of grid_cells.
                3) jlist j-index of grid_cells.
                4) wghtslist weight values of i,j index of grid_cells.
        """
        if n_jobs == -1:
            n_jobs = os.cpu_count()

        # Buffer polygons with self-intersections
        tstrt = time.perf_counter()
        df1 = _make_valid(source_df)
        df2 = _make_valid(target_df)
        tend = time.perf_counter()
        print(f"Validate polygons finished in {tend - tstrt:0.4f} seconds")

        # Chunk the largest, ship the smallest in full
        to_chunk, df_full = self.get_chunks(df1, df2)

        # Spatial index query: Reindex on positional IDs
        to_workers = _chunk_dfs(
            gpd.GeoSeries(to_chunk.geometry.values, crs=to_chunk.crs),
            gpd.GeoSeries(df_full.geometry.values, crs=df_full.crs),
            n_jobs,
        )

        worker_out = self.get_ids_parallel(n_jobs, to_workers)
        ids_src, ids_tgt = np.concatenate(worker_out).T

        # Intersection + area calculation
        chunks_to_intersection = _chunk_polys(
            np.vstack([ids_src, ids_tgt]).T, df1.geometry, df2.geometry, n_jobs
        )
        worker_out = self.get_areas_and_intersections_parallel(
            n_jobs, chunks_to_intersection
        )
        areas = np.concatenate([item[0] for item in worker_out])
        inter_geom = np.concatenate([item[1] for item in worker_out])

        print("Processing intersections for output.")
        inter_sect = df2.iloc[ids_tgt, :].set_geometry(inter_geom)
        weights = areas.astype(float) / df2.geometry[ids_tgt].area
        inter_sect["weights"] = weights

        return (
            df2[self.poly_idx].iloc[ids_tgt].values.astype(object).tolist(),
            df1.i_index.iloc[ids_src].values.astype(int).tolist(),
            df1.j_index.iloc[ids_src].values.astype(int).tolist(),
            (areas.astype(float) / df2.geometry[ids_tgt].area).tolist(),
            inter_sect,
        )

    def get_areas_parallel(self, n_jobs, chunks_to_intersection):
        """Get poly-to-poly intersections."""
        with parallel_backend("loky", inner_max_num_threads=1):
            worker_out = Parallel(n_jobs=n_jobs)(
                delayed(_area_on_chunk)(*chunk_pair)
                for chunk_pair in chunks_to_intersection
            )
        return worker_out

    def get_areas_and_intersections_parallel(self, n_jobs, chunks_to_intersection):
        """Get poly-to-poly intersections."""
        with parallel_backend("loky", inner_max_num_threads=1):
            worker_out = Parallel(n_jobs=n_jobs)(
                delayed(_intersect_area_on_chunk)(*chunk_pair)
                for chunk_pair in chunks_to_intersection
            )
        return worker_out

    def get_ids_parallel(self, n_jobs, to_workers):
        """Get poly-to-poly intersection ids."""
        with parallel_backend("loky", inner_max_num_threads=1):
            worker_out = Parallel(n_jobs=n_jobs)(
                delayed(_index_n_query)(*chunk_pair) for chunk_pair in to_workers
            )
        return worker_out

    def get_chunks(self, df1, df2):
        """Chunk dataframes."""
        to_chunk = df1
        df_full = df2
        return to_chunk, df_full

    def get_weights_values(self, poly, poly_idx, grid_cells, results):
        """Extract values from Tobler result for weight file."""
        coo = results.tocoo().toarray()
        weight_index = np.transpose((coo > 0).nonzero())
        weight_array = coo[np.where(coo > 0)]
        idx_array = poly.loc[:, poly_idx].values
        poly_idx_array = np.array(
            [
                idx_array[weight_index[:, 1][i]]
                for i in np.arange(weight_index[:, 1].shape[0])
            ]
        ).astype(object)
        i_array = grid_cells.loc[:, "i_index"].values[weight_index[:, 0]].astype(int)
        j_array = grid_cells.loc[:, "j_index"].values[weight_index[:, 0]].astype(int)
        w_array = weight_array / poly.geometry.area[weight_index[:, 1]].values.astype(
            float
        )
        return (
            poly_idx_array.tolist(),
            i_array.tolist(),
            j_array.tolist(),
            w_array.tolist(),
        )


def _chunk_dfs(geoms_to_chunk, geoms_full, n_jobs):
    """Chunk dataframes for parallel processing."""
    chunk_size = geoms_to_chunk.shape[0] // n_jobs + 1
    for i in range(n_jobs):
        start = i * chunk_size
        yield geoms_to_chunk.iloc[start : start + chunk_size], geoms_full


def _index_n_query(geoms1, geoms2):
    """Get geom ids for parallel processing."""
    # Pick largest for STRTree, query_bulk the smallest

    # Build tree + query
    qry_polyids, tree_polyids = geoms1.sindex.query_bulk(geoms2, predicate="intersects")
    # Remap IDs to global
    large_global_ids = geoms1.iloc[tree_polyids].index.values
    small_global_ids = geoms2.iloc[qry_polyids].index.values

    return np.array([large_global_ids, small_global_ids]).T


def _chunk_polys(id_pairs, geoms_left, geoms_right, n_jobs):
    """Chunk polys for parallel processing."""
    chunk_size = id_pairs.shape[0] // n_jobs + 1
    for i in range(n_jobs):
        start = i * chunk_size
        chunk1 = geoms_left.values.data[id_pairs[start : start + chunk_size, 0]]
        chunk2 = geoms_right.values.data[id_pairs[start : start + chunk_size, 1]]
        yield chunk1, chunk2


def _area_on_chunk(geoms1, geoms2):
    """Get intersection areas."""
    return pygeos.area(pygeos.intersection(geoms1, geoms2))


def _intersect_area_on_chunk(geoms1, geoms2):
    """Get intersection areas."""
    f_intersect = pygeos.intersection(geoms1, geoms2)
    return pygeos.area(f_intersect), f_intersect


def _make_valid(df: gpd.GeoDataFrame):
    """Make in-valid geometries valid.

    Based on the method in Shapely and slightly modified for use here:

    Copyright (c) 2007, Sean C. Gillies. 2019, Casper van der Wel. 2007-2022,
    Shapely Contributors. All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    Args:
        df (gpd.GeoDataFrame): _description_

    Returns:
        _type_: _description_
    """
    polys = ["Polygon", "MultiPolygon"]
    df = df.copy()
    if df.geom_type.isin(polys).all():
        mask = ~df.geometry.is_valid
        col = df._geometry_column_name
        df.loc[mask, col] = df.loc[mask, col].buffer(0)
    return df



"""Calculate zonal methods."""
import time
from datetime import datetime
from pathlib import Path
from typing import Literal
from typing import Optional

import pandas as pd

from gdptools.agg.zonal_engines import ZonalEngineParallel
from gdptools.agg.zonal_engines import ZonalEngineSerial
from gdptools.data.user_data import UserData

ZONAL_ENGINES = Literal["serial", "parallel"]
ZONAL_WRITERS = Literal["csv"]


class ZonalGen:
    """Class for aggregating zonal statistics."""

    def __init__(
        self,
        user_data: UserData,
        zonal_engine: ZONAL_ENGINES,
        zonal_writer: ZONAL_WRITERS,
        out_path: str,
        file_prefix: str,
        append_date: Optional[bool] = False,
    ) -> None:
        """__init__ Initialize ZonalGen class.

        _extended_summary_

        Args:
            user_data (UserData): _description_
            zonal_engine (ZONAL_ENGINES): _description_
            zonal_writer (ZONAL_WRITERS): _description_
            out_path (str): _description_
            file_prefix (str): _description_
            append_date (Optional[bool], optional): _description_. Defaults to False.

        Raises:
            FileNotFoundError: _description_
            TypeError: _description_
        """
        self._user_data = user_data
        self._zonal_engine = zonal_engine
        self._zonal_writer = zonal_writer
        self._out_path = Path(out_path)
        if not self._out_path.exists():
            raise FileNotFoundError(f"Path: {self._out_path} does not exist")
        self._file_prefix = file_prefix
        self._append_date = append_date
        if self._append_date:
            self._fdate = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            self._fname = f"{self._fdate}_{self._file_prefix}"
        else:
            self._fname = f"{self._file_prefix}"

        if self._zonal_engine == "serial":
            self.agg = ZonalEngineSerial
        elif self._zonal_engine == "parallel":
            self.agg = ZonalEngineParallel
        else:
            raise TypeError(f"agg_engine: {self._zonal_engine} not in {ZONAL_ENGINES}")

    def calculate_zonal(self, categorical: Optional[bool] = False) -> pd.DataFrame:
        """calculate_zonal Calculates zonal statistics.

        _extended_summary_

        Args:
            categorical (Optional[bool], optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: _description_
        """
        tstrt = time.perf_counter()
        stats = self.agg().calc_zonal_from_aggdata(
            user_data=self._user_data, categorical=categorical
        )
        if self._zonal_writer == "csv":
            fullpath = self._out_path / f"{self._fname}.csv"
            stats.to_csv(fullpath, sep=",")
        tend = time.perf_counter()
        print(
            f"Total time for serial zonal stats calculation {tend - tstrt:0.4f} seconds"
        )
        return stats
        # elif self._zonal_writer == "feather":
        #     fullpath = self._out_path / f"{self._fname}"
        #     stats.to_feather(path=fullpath, )



"""Data classes used in aggregation."""
from dataclasses import dataclass

import geopandas as gpd


@dataclass(repr=True)
class WeightData(object):
    """Simple dataclass for tranferring prepared user data to the CalcWeightEngine."""

    feature: gpd.GeoDataFrame
    id_feature: str
    grid_cells: gpd.GeoDataFrame



"""Aggregation engines."""
import time
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Tuple
from typing import Union

import geopandas as gpd
import numpy as np
from numpy.typing import NDArray

from gdptools.agg.stats_methods import StatsMethod
from gdptools.data.agg_gen_data import AggData
from gdptools.data.user_data import UserData
from gdptools.utils import _get_default_val
from gdptools.utils import _get_interp_array
from gdptools.utils import _get_wieght_df


class AggEngine(ABC):
    """Abstract aggregation class."""

    def calc_agg_from_dictmeta(
        self,
        user_data: UserData,
        weights: Union[str, gpd.GeoDataFrame],
        stat: StatsMethod,
    ) -> Tuple[gpd.GeoDataFrame, List[NDArray[np.double]]]:
        """Abstract Base Class for calculating aggregations from dictionary metadata.

        Args:
            user_data (UserData): _description_
            weights (Union[str, gpd.GeoDataFrame]): _description_
            stat (StatsMethod): _description_

        Returns:
            Tuple[gpd.GeoDataFrame, List[NDArray[np.double]]]: _description_
        """
        self.usr_data = user_data
        self.id_feature = user_data.get_feature_id()
        self.vars = user_data.get_vars()
        self.stat = stat
        self.period = None
        self.wghts = _get_wieght_df(weights, self.id_feature)

        return self.agg_w_weights()

    @abstractmethod
    def agg_w_weights(self) -> Tuple[gpd.GeoDataFrame, List[NDArray[np.double]]]:
        """Abstract method for calculating weights."""
        pass


class SerialAgg(AggEngine):
    """SerialAgg data by feature and time period."""

    def get_period_from_ds(self, data: AggData) -> List[str]:
        """Get starting and ending time string from previously subsetted Dataset.

        Args:
            data (AggData): _description_

        Returns:
            List[str]: _description_
        """
        # tname = list(data.cat_param.values())[0]["T_name"]
        tname = data.cat_param.T_name
        tstrt = data.da.coords[tname].values[0]
        tend = data.da.coords[tname].values[-1]
        return [tstrt, tend]

    def agg_w_weights(
        self,
    ) -> Tuple[dict[str, AggData], gpd.GeoDataFrame, List[NDArray[np.double]]]:
        """Standard aggregate method.

        Returns:
            Tuple[List[AggData], gpd.GeoDataFrame, List[NDArray[np.double]]]:
                _description_
        """
        # ds_time = self.ds.coords[list(self.param_dict.values())[0]["T_name"]].values
        # date_bracket = np.array_split(ds_time, self.numdiv)
        # print(date_bracket)
        # date_bracket = list(_date_range(self.period[0], self.period[1], self.numdiv))
        r_gdf = []
        r_vals = []
        r_agg_dict = {}
        avars = self.usr_data.get_vars()
        for index, key in enumerate(avars):
            print(f"Processing: {key}")
            tstrt = time.perf_counter()
            agg_data: AggData = self.usr_data.prep_agg_data(key=key)
            tend = time.perf_counter()
            print(f"    Data prepped for aggregation in {tend-tstrt:0.4f} seconds")
            tstrt = time.perf_counter()
            newgdf, nvals = self.calc_agg(key=key, data=agg_data)
            tend = time.perf_counter()
            print(f"    Data aggregated in {tend -tstrt:0.4f} seconds")
            if index == 0:
                # all new GeoDataFrames will be the same so save and return only one.
                r_gdf.append(newgdf)
            r_vals.append(nvals)
            r_agg_dict[key] = agg_data
        return r_agg_dict, r_gdf[0], r_vals

    def calc_agg(
        self: "SerialAgg", key: str, data: AggData
    ) -> Tuple[gpd.GeoDataFrame, NDArray[np.double]]:
        """Calculate aggregation.

        Args:
            key (str): _description_
            data (AggData): _description_

        Returns:
            Tuple[gpd.GeoDataFrame, NDArray[np.double]]: _description_
        """
        cp = data.cat_param
        period = self.get_period_from_ds(data=data)
        gdf = data.feature
        gdf.reset_index(drop=True, inplace=True)
        gdf = gdf.sort_values(data.id_feature).dissolve(by=data.id_feature)
        geo_index = np.asarray(gdf.index, dtype=type(gdf.index.values[0]))
        n_geo = len(geo_index)
        unique_geom_ids = self.wghts.groupby(self.id_feature)
        t_name = cp.T_name
        selection = {t_name: slice(period[0], period[1])}
        da = data.da.sel(**selection)
        # print(da)
        nts = len(da.coords[t_name].values)
        native_dtype = da.dtype
        # gdptools will handle floats and ints - catch if gridded type is different
        try:
            dfval = _get_default_val(native_dtype=native_dtype)
        except TypeError as e:
            print(e)

        val_interp = _get_interp_array(
            n_geo=n_geo, nts=nts, native_dtype=native_dtype, default_val=dfval
        )

        mdata = np.ma.masked_array(da.values, np.isnan(da.values))

        for i in np.arange(len(geo_index)):
            try:
                weight_id_rows = unique_geom_ids.get_group(str(geo_index[i]))
            except KeyError:
                continue
            tw = weight_id_rows.wght.values
            i_ind = np.array(weight_id_rows.i.values).astype(int)
            j_ind = np.array(weight_id_rows.j.values).astype(int)

            val_interp[:, i] = self.stat(
                array=mdata[:, i_ind, j_ind], weights=tw, def_val=dfval
            ).get_stat()

        return gdf, val_interp


# class ParallelAgg(AggEngine):
